"""Cross-platform reimplementations of common coreutils.

Pure-Python versions of ``cat``, ``echo``, ``pwd``, ``tee``, ``tty``,
``ulimit``, ``umask``, ``uname``, ``uptime``, ``which``, ``yes`` and
related helpers. Used so that xonsh scripts behave identically across
POSIX and Windows.
"""
