"""Web-based configuration UI for the xonsh shell.

Bundles a small HTTP server and HTML/JS front-end that lets users
browse and configure xontribs, prompts, colors and other xonsh
settings from a browser. Launched via ``xonfig web``.
"""
