"""Tab completion providers for the xonsh shell.

Bundles the built-in completers for Python expressions, file paths,
commands, bash completions, man pages, imports, environment variables
and more.
"""
